#include "Message.h"
#include <iostream>


const QString Message::dateTimeFormat = "hh:mm:ss dd.MM.yyyy";


Message Message::from_json_byte(const QByteArray &array)
{
    auto e = QJsonParseError();
    auto msg = Message(QJsonDocument::fromJson(array, &e).object());
    if (e.error != QJsonParseError::NoError) {
        throw std::runtime_error(e.errorString().toStdString());
    }
    /*if (msg.get_type() == Type::INFO) {
        std::cout << "Creating info, initial: |" << array.toStdString() << "| created: " << msg.get_raw_string().toStdString() << "@" << std::endl;

    }*/

    if (!msg.m_scheme.contains("type")) {
        throw std::runtime_error("Invalid message type, field \"type\" is missing");
    }

    if (!msg.m_scheme.contains("time")) {
        throw std::runtime_error("Invalid message, field \"time\" is missing");
    }

    auto time = msg.m_scheme["time"].toString();

    if (time.length() != dateTimeFormat.length())
    {
        throw std::runtime_error("Invalid message time format, required: " + dateTimeFormat.toStdString()
        + "\n actual: " + time.toStdString());
    }

    for (int i = 0; i < dateTimeFormat.size(); ++i)
    {
        if ((dateTimeFormat[i] == ':' || dateTimeFormat[i] == ' ' || dateTimeFormat[i] == '.')
            && dateTimeFormat[i] != time[i])
        {
            throw std::runtime_error("Invalid message time format, required: " + dateTimeFormat.toStdString()
                                     + "\n actual: " + time.toStdString());
        }
    }

    switch (static_cast<Type>(msg.m_scheme["type"].toInt()))
    {
        case Type::TEXT:
            if (!msg.m_scheme.contains("sender") || !msg.m_scheme.contains("recipient")
                || !msg.m_scheme.contains("message") || msg.m_scheme.keys().length() != 2 + 3)
                throw std::runtime_error("Text message has invalid fields");
            break;
        case Type::LOGIN:
        case Type::REGISTER:
            if (!msg.m_scheme.contains("password")
                || !msg.m_scheme.contains("sender") || msg.m_scheme.keys().length() != 2 + 2)
                throw std::runtime_error("Login/register message has invalid fields");
            break;
        case Type::LOGOUT:
            if (!msg.m_scheme.contains("sender") || msg.m_scheme.keys().length() != 2 + 1)
                throw std::runtime_error("Logout message has invalid fields");
            break;
        case Type::INFO:
            if (!msg.m_scheme.contains("initial_message") || !msg.m_scheme.contains("recipient")
                || !msg.m_scheme.contains("success") || !msg.m_scheme.contains("info")
                || msg.m_scheme.keys().length() != 2 + 4)
                throw std::runtime_error("Login/register message has invalid fields");
            break;
        case Type::FRIEND_REQUEST:
            if (!msg.m_scheme.contains("sender") || !msg.m_scheme.contains("recipient")
                || msg.m_scheme.keys().length() != 2 + 2)
                throw std::runtime_error("Friend request message has invalid fields");
            break;
        case Type::UNFRIEND_REQUEST:
            if (!msg.m_scheme.contains("sender") || !msg.m_scheme.contains("recipient")
                || msg.m_scheme.keys().length() != 2 + 2)
                throw std::runtime_error("UNFRIEND_REQUEST message has invalid fields");
            break;
        case Type::FRIEND_REQUEST_RESPONSE:
            if (!msg.m_scheme.contains("sender") || !msg.m_scheme.contains("recipient")
                || !msg.m_scheme.contains("accepted") || msg.m_scheme.keys().length() != 2 + 3)
                throw std::runtime_error("Response to friend request has invalid fields");
            break;
        case Type::CONTACT_LIST_REQUEST:
            if (!msg.m_scheme.contains("sender") || msg.m_scheme.keys().length() != 2 + 1)
                throw std::runtime_error("CONTACT_LIST_REQUEST message has invalid fields");
            break;
        case Type::CONTACT_LIST_RESPONSE:
            if (!msg.m_scheme.contains("friends") || msg.m_scheme.keys().length() != 2 + 1)
                throw std::runtime_error("CONTACT_LIST_RESPONSE has invalid fields");
            break;
        default:
            throw std::runtime_error("Invalid message type");
    }

    return msg;
}


/**
 * NOTE: Qt's support for json does not preserve the order of the insertion of key-value pairs to a QJsonObject,
 * but keeps them alphabetically ordered
 */
Message Message::create_text_message(const QString &sender, const QString &recipient, const QString &msg)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::TEXT) },
       { QStringLiteral("sender"), sender },
       { QStringLiteral("recipient"), recipient },
       { QStringLiteral("message"), msg }
    }));
}


Message Message::create_login_message(const QString &sender, const QString &password)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::LOGIN) },
       { QStringLiteral("sender"), sender },
       { QStringLiteral("password"), password }
    }));
}


Message Message::create_register_message(const QString &sender, const QString &password)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::REGISTER)},
       { QStringLiteral("sender"), sender },
       { QStringLiteral("password"), password }
    }));
}


Message Message::create_friend_request(const QString& sender, const QString& recipient)
{
    return Message(QJsonObject({
        { QStringLiteral("type"), static_cast<int>(Type::FRIEND_REQUEST) },
        { QStringLiteral("sender"), sender },
        { QStringLiteral("recipient"), recipient }
    }));
}


Message Message::create_unfriend_request(const QString &sender, const QString &friend_)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::UNFRIEND_REQUEST) },
       { QStringLiteral("sender"), sender },
       { QStringLiteral("recipient"), friend_ }
    }));
}


Message Message::create_logout_message(const QString &sender)
{
    return Message(QJsonObject({
        { QStringLiteral("type"), static_cast<int>(Type::LOGOUT) },
        { QStringLiteral("sender"), sender }}));
}


Message Message::create_friend_request_response(const QString &sender, const QString &recipient, bool accepted)
{
    return Message(QJsonObject({
        { QStringLiteral("type"), static_cast<int>(Type::FRIEND_REQUEST_RESPONSE) },
        { QStringLiteral("sender"), sender },
        { QStringLiteral("recipient"), recipient },
        { QStringLiteral("accepted"), accepted }
    }));
}


Message Message::create_info_message(const QString &recipient, const Message &initial_msg, bool success, QString info)
{
    return Message(QJsonObject({
        { QStringLiteral("type"), static_cast<int>(Type::INFO) },
        { QStringLiteral("recipient"), recipient },
        { QStringLiteral("initial_message"), initial_msg.get_raw_string() },
        { QStringLiteral("success"), success },
        { QStringLiteral("info"), info }
    }));
}



Message Message::create_contact_list_request(const QString &sender)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::CONTACT_LIST_REQUEST) },
       { QStringLiteral("sender"), sender }
    }));
}

Message Message::create_contact_list_response(const QJsonArray &friends)
{
    return Message(QJsonObject({
       { QStringLiteral("type"), static_cast<int>(Type::CONTACT_LIST_RESPONSE) },
       { QStringLiteral("friends"), friends }
    }));
}


bool Message::equals_without_time(const Message &other) const
{
    auto temp = m_scheme;
    auto temp2 = other.m_scheme;
    //std::cout << "Came: " + QString(QJsonDocument(temp).toJson()).toStdString() + " correct: " + QString(QJsonDocument(temp2).toJson()).toStdString();

    temp.remove("time");
    temp2.remove("time");
    //temp.insert("time", other.get_time().toString());
    return temp == temp2;
}


